//-----------------------------------------------------------------------------
// File: objects.cpp
//
// Desc: Example code showing how to rotating two objects independantly 
//       in a scene and translating and rotating a camera
//
//       Note: This code uses the D3D Framework helper library.
//
// Copyright (c) 1997-1999 Microsoft Corporation. All rights reserved.
// Partly copyright (c) 1999 - 2001 Wolfgang Engel wolf@direct3d.net
//-----------------------------------------------------------------------------
#define STRICT
#define D3D_OVERLOADS
#include <stdio.h>
#include <math.h>
#include <time.h>
#include "D3DTextr.h"
#include "D3DUtil.h"
#include "D3DMath.h"
#include "D3DApp.h"
#include "D3DFile.h"
#include "dinput.h"


//-----------------------------------------------------------------------------
// Name: struct Object
// Desc: Structure for holding data for each object
//-----------------------------------------------------------------------------
struct Object
{
    CD3DFile*   m_pFile;
    D3DVECTOR   vLoc;
    FLOAT       fYaw, fPitch, fRoll; 
    D3DMATRIX   matLocal;
};



//-----------------------------------------------------------------------------
// Name: class CMyD3DApplication
// Desc: Application class. The base class provides just about all the
//       functionality we want, so we're just supplying stubs to interface with
//       the non-C++ functions of the app.
//-----------------------------------------------------------------------------
class CMyD3DApplication : public CD3DApplication
{
    D3DVERTEX  m_pvObjectVertices[16];
    WORD       m_pwObjectIndices[30];
    Object     m_pObjects[2];

    
	FLOAT m_fStartTimeKey,     // Time reference for calculations
		  m_fTimeElapsed;

    static HRESULT ConfirmDevice( DDCAPS* pddDriverCaps, 
                                  D3DDEVICEDESC7* pd3dDeviceDesc );
protected:
    HRESULT OneTimeSceneInit();
    HRESULT InitDeviceObjects();
    HRESULT DeleteDeviceObjects();
    HRESULT Render();
    HRESULT FrameMove( FLOAT fTimeKey );
    HRESULT FinalCleanup();

public:
	LPDIRECTINPUT7       g_Keyboard_pDI;            // The DInput object
	LPDIRECTINPUTDEVICE2 g_Keyboard_pdidDevice2;    // The DIDevice2 interface
    
	CMyD3DApplication();

	// dinput methods
	HRESULT CreateDInput( HWND hWnd );
	VOID DestroyInputDevice();
	VOID UpdateControls();
	HRESULT CreateInputDevice( HWND hWnd, 
							   LPDIRECTINPUT7       pDI,
							   LPDIRECTINPUTDEVICE2 pDIdDevice, 
							   GUID guidDevice,
							   const DIDATAFORMAT* pdidDataFormat, 
							   DWORD dwFlags );
	VOID DestroyDInput();

	// utitlity methods
    HWND Get_hWnd()	{ return m_hWnd; };

};



//-----------------------------------------------------------------------------
// Name: WinMain()
// Desc: Entry point to the program. Initializes everything, and goes into a
//       message-processing loop. Idle time is used to render the scene.
//-----------------------------------------------------------------------------
INT WINAPI WinMain( HINSTANCE hInst, HINSTANCE, LPSTR strCmdLine, INT )
{
    CMyD3DApplication d3dApp;

    if( FAILED( d3dApp.Create( hInst, strCmdLine ) ) )
        return 0;

	// Create the DInput object
	if( FAILED(d3dApp.CreateDInput( d3dApp.Get_hWnd() ) ) )
        return 0;
   
    // Create a keyboard device
    if( FAILED(d3dApp.CreateInputDevice( d3dApp.Get_hWnd(), 
										 d3dApp.g_Keyboard_pDI, 
										 d3dApp.g_Keyboard_pdidDevice2, 
										 GUID_SysKeyboard, 
										 &c_dfDIKeyboard,
										 DISCL_NONEXCLUSIVE | DISCL_FOREGROUND)))
	    return 0;

	return d3dApp.Run();
}

//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Application constructor. Sets attributes for the app.
//-----------------------------------------------------------------------------
CMyD3DApplication::CMyD3DApplication()
                  :CD3DApplication()
{
    m_strWindowTitle  = TEXT("Building Worlds with X files");
    m_bAppUseZBuffer  = TRUE;
    m_bAppUseStereo   = FALSE;
    m_bShowStats      = TRUE;
    m_fnConfirmDevice = ConfirmDevice;

}


//-----------------------------------------------------------------------------
// Name: OneTimeSceneInit()
// Desc: Called during initial app startup, this function performs all the
//       permanent initialization.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::OneTimeSceneInit()
{
	// yellow object
	m_pObjects[0].vLoc   = D3DVECTOR(-1.0f, 0.0f, 0.0f);
    m_pObjects[0].fYaw   = 0.0f;
    m_pObjects[0].fPitch = 0.0f;
    m_pObjects[0].fRoll  = 0.0f;

	// red object
	m_pObjects[1].vLoc   = D3DVECTOR(1.0f, 0.0f, 0.0f);
    m_pObjects[1].fYaw   = 0.0f;
    m_pObjects[1].fPitch = 0.0f;
    m_pObjects[1].fRoll  = 0.0f;


    HRESULT hr;

    // yellow object
    m_pObjects[0].m_pFile  = new CD3DFile();
    
	hr = m_pObjects[0].m_pFile->Load ("boidy2.x");

    if( FAILED(hr) )
	{
        MessageBox( NULL, TEXT("Error loading boidy.x file"),
                    TEXT("Animation"), MB_OK|MB_ICONERROR);
		return E_FAIL;
	}
    
    // get the X file matrix
    CD3DFileObject* pObject = m_pObjects[0].m_pFile->FindObject( "BOID_Root" );

    if( pObject )
    {
        m_pObjects[0].matLocal = *pObject->GetMatrix();
	}

    // red object
    m_pObjects[1].m_pFile  = new CD3DFile();

    // Cube will be scaled in FrameMove () with z
	hr = m_pObjects[1].m_pFile->Load ("boidr2.x");

    if( FAILED(hr) )
	{
        MessageBox( NULL, TEXT("Error loading boidr.x file"),
                    TEXT("Animation"), MB_OK|MB_ICONERROR);
		return E_FAIL;
	}
    
    // get the X file matrix
    CD3DFileObject* pObject2 = m_pObjects[1].m_pFile->FindObject( "BOID_Root" );

    if( pObject2 )
    {
        m_pObjects[1].matLocal = *pObject2->GetMatrix();
	}

  return S_OK;
}


//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Called once per frame, the call is the entry point for 3d
//       rendering. This function sets up render states, clears the
//       viewport, and renders the scene.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::Render()
{

    // Clear the viewport
    m_pd3dDevice->Clear( 0, NULL, D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER,
                         0x00000000, 1.0f, 0L );

    // Begin the scene
    if( FAILED( m_pd3dDevice->BeginScene() ) )
        return S_OK; // Don't return a "fatal" error

		if (m_pObjects[0].m_pFile)
			m_pObjects[0].m_pFile->Render( m_pd3dDevice );

        // Apply the object's local matrix
        m_pd3dDevice->SetTransform(D3DTRANSFORMSTATE_WORLD, &m_pObjects[0].matLocal);

		if (m_pObjects[1].m_pFile)
			m_pObjects[1].m_pFile->Render( m_pd3dDevice );

        // Apply the object's local matrix
        m_pd3dDevice->SetTransform(D3DTRANSFORMSTATE_WORLD, &m_pObjects[1].matLocal);


    // End the scene.
    m_pd3dDevice->EndScene();

    return S_OK;
}


//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::InitDeviceObjects()
{
    D3DMATERIAL7 mtrl;
    D3DUtil_InitMaterial( mtrl, 1.0f, 1.0f, 1.0f );
    m_pd3dDevice->SetMaterial( &mtrl );

    m_pd3dDevice->SetRenderState( D3DRENDERSTATE_ZENABLE, TRUE); //D3DZB_USEW );
	
/*	// wireframe object
    m_pd3dDevice->SetRenderState( D3DRENDERSTATE_FILLMODE, D3DFILL_WIREFRAME );

    // no backface culling
	m_pd3dDevice->SetRenderState(D3DRENDERSTATE_CULLMODE, 
                            D3DCULL_NONE);
*/

    // Set the transform matrices
    D3DVIEWPORT7 vp;
    m_pd3dDevice->GetViewport(&vp);
    FLOAT fAspect = ((FLOAT)vp.dwHeight) / vp.dwWidth;

    D3DMATRIX matProj;
    D3DUtil_SetProjectionMatrix( matProj, 0.75f, fAspect, 1.0f, 1000.0f );
    m_pd3dDevice->SetTransform( D3DTRANSFORMSTATE_PROJECTION, &matProj );


    // Set up the lights
    m_pd3dDevice->SetRenderState( D3DRENDERSTATE_AMBIENT, 0x0c0c0c0c);

    if( m_pDeviceInfo->ddDeviceDesc.dwVertexProcessingCaps &
                       D3DVTXPCAPS_DIRECTIONALLIGHTS )
    {
        D3DLIGHT7 light;

        if( m_pDeviceInfo->ddDeviceDesc.dwMaxActiveLights > 0 )
        {
            D3DUtil_InitLight( light, D3DLIGHT_DIRECTIONAL, 0.5f, -1.0f, 0.3f );
            m_pd3dDevice->SetLight( 0, &light );
            m_pd3dDevice->LightEnable( 0, TRUE );
        }

        if( m_pDeviceInfo->ddDeviceDesc.dwMaxActiveLights > 1 )
        {
            D3DUtil_InitLight( light, D3DLIGHT_DIRECTIONAL, 0.5f, 1.0f, 1.0f );
            light.dcvDiffuse.r = 0.5f;
            light.dcvDiffuse.g = 0.5f;
            light.dcvDiffuse.b = 0.5f;
            m_pd3dDevice->SetLight( 1, &light );
            m_pd3dDevice->LightEnable( 1, TRUE );
        }

        m_pd3dDevice->SetRenderState( D3DRENDERSTATE_LIGHTING, TRUE );
    }

    return S_OK;
}

//-----------------------------------------------------------------------------
// Name: FinalCleanup()
// Desc: Called before the app exits, this function gives the app the chance
//       to cleanup after itself.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::FinalCleanup()
{
	// release keyboard stuff
	DestroyInputDevice();

    return S_OK;
}


//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exitting, or the device is being changed,
//       this function deletes any device dependant objects.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::DeleteDeviceObjects()
{

   return S_OK;
}

//-----------------------------------------------------------------------------
// Name: ConfirmDevice()
// Desc: Called during device intialization, this code checks the device
//       for some minimum set of capabilities
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::ConfirmDevice( DDCAPS* pddDriverCaps,
                                          D3DDEVICEDESC7* pd3dDeviceDesc )
{
    return S_OK;
}

//-----------------------------------------------------------------------------
// Name: CreateDInput()
// Desc: Initialize the DirectInput objects.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::CreateDInput( HWND hWnd )
{
	// keyboard
	if( FAILED(DirectInputCreateEx( (HINSTANCE)GetWindowLong( hWnd, GWL_HINSTANCE ),
                                   DIRECTINPUT_VERSION, 
								   IID_IDirectInput7, 
								   (LPVOID*) &g_Keyboard_pDI, NULL) ) )
        return 0;
	
    return S_OK;
}


//-----------------------------------------------------------------------------
// Name: CreateInputDevice()
// Desc: Create a DirectInput device.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::CreateInputDevice( HWND hWnd, 
								   LPDIRECTINPUT7       pDI,	
								   LPDIRECTINPUTDEVICE2 pDIdDevice, 
								   GUID guidDevice,
								   const DIDATAFORMAT* pdidDataFormat, DWORD dwFlags )
{

    // Obtain an interface to the input device
    if( FAILED( pDI->CreateDeviceEx( guidDevice, IID_IDirectInputDevice2,
		                               (VOID**)&pDIdDevice, NULL ) ) )
        return 0;

    // Set the device data format
    if( FAILED( pDIdDevice->SetDataFormat( pdidDataFormat ) ) )
        return 0;

    // Set the cooperativity level
    if( FAILED( pDIdDevice->SetCooperativeLevel( hWnd, dwFlags ) ) )
        return 0;

	if(guidDevice == GUID_SysKeyboard)
		g_Keyboard_pdidDevice2 = pDIdDevice;

    return S_OK;
}

//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Called once per frame, the call is the entry point for animating
//       the scene.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::FrameMove (FLOAT fTimeKey)
{
    
	BYTE diks[256]; // DInput keyboard state buffer
    ZeroMemory(diks, sizeof(diks));

	if (FAILED(g_Keyboard_pdidDevice2->GetDeviceState( sizeof(diks), &diks )))
	{
		g_Keyboard_pdidDevice2->Acquire();
		if (FAILED(g_Keyboard_pdidDevice2->GetDeviceState( sizeof(diks), &diks )))
			return 0;
	}

	// timing code: 
	// the object should move/rotate in the same speed 
	// at every possible fps
	const cTimeScale = 5;
	
	// calculate elapsed time
	m_fTimeElapsed=(fTimeKey-m_fStartTimeKey) * cTimeScale;
	
	// store last time
	m_fStartTimeKey=fTimeKey;

	// timing variables
	FLOAT fTimedvar = 1.0f * m_fTimeElapsed;
	FLOAT fTimedvarcamera = 0.3f * m_fTimeElapsed;


	// yellow object
	if (diks[DIK_J] &&0x80) // j key
		m_pObjects[0].fRoll -= fTimedvar;

    if (diks[DIK_L] &&0x80)  // l key
		m_pObjects[0].fRoll += fTimedvar;

	if (diks[DIK_I] &&0x80) // i key
		m_pObjects[0].fPitch -= fTimedvar;

    if (diks[DIK_K] &&0x80) // k key       
		m_pObjects[0].fPitch += fTimedvar;


    D3DMATRIX matWorld;

    // object matrix for yellow object
	D3DUtil_SetTranslateMatrix( matWorld, m_pObjects[0].vLoc);

    D3DMATRIX matTemp, matRotateX, matRotateY, matRotateZ;
    D3DUtil_SetRotateYMatrix( matRotateY, -m_pObjects[0].fYaw );
    D3DUtil_SetRotateXMatrix( matRotateX, -m_pObjects[0].fPitch );
    D3DUtil_SetRotateZMatrix( matRotateZ, -m_pObjects[0].fRoll );
    D3DMath_MatrixMultiply( matTemp, matRotateX, matRotateY );
    D3DMath_MatrixMultiply( matTemp, matRotateZ, matTemp );
    D3DMath_MatrixMultiply( matWorld, matTemp, matWorld );

    m_pObjects[0].matLocal = matWorld;

    
	// red object
    if (diks[DIK_D] &&0x80)   // Key d
		m_pObjects[1].fRoll -= fTimedvar;

    if (diks[DIK_A] &&0x80)   // Key a
		m_pObjects[1].fRoll += fTimedvar;
	
	if (diks[DIK_S] &&0x80)   // Key s
		m_pObjects[1].fPitch -= fTimedvar;

    if (diks[DIK_W] &&0x80)   // Key w     
		m_pObjects[1].fPitch += fTimedvar;

	// object matrix for red object
	D3DUtil_SetTranslateMatrix( matWorld, m_pObjects[1].vLoc );

    D3DUtil_SetRotateYMatrix( matRotateY, -m_pObjects[1].fYaw );
    D3DUtil_SetRotateXMatrix( matRotateX, -m_pObjects[1].fPitch );
    D3DUtil_SetRotateZMatrix( matRotateZ, -m_pObjects[1].fRoll );
    D3DMath_MatrixMultiply( matTemp, matRotateX, matRotateY );
    D3DMath_MatrixMultiply( matTemp, matRotateZ, matTemp );
    D3DMath_MatrixMultiply( matWorld, matTemp, matWorld );

    m_pObjects[1].matLocal = matWorld;

	//************************************************************
	// camera stuff
	//************************************************************
	static D3DVECTOR vLook=D3DVECTOR(0.0f,0.0f,1.0);
	static D3DVECTOR vUp=D3DVECTOR(0.0f,1.0f,0.0f);
	static D3DVECTOR vRight=D3DVECTOR(1.0f,0.0f,0.0f);
	static D3DVECTOR vPos=D3DVECTOR(0.0f,0.0f,-5.0f);
	FLOAT fPitch,fYaw,fRoll;
	fPitch = fYaw = fRoll = 0.0f;

	FLOAT fspeed= fTimedvar;

	// fPitch
    if (diks[DIK_UP] &&0x80) 
		fPitch=-fTimedvarcamera;

    if (diks[DIK_DOWN] &&0x80)
		fPitch=+fTimedvarcamera;

	// fYaw
    if (diks[DIK_C] &&0x80) // c key
		fYaw=-fTimedvarcamera;

    if (diks[DIK_X] &&0x80) // x key
		fYaw=+fTimedvarcamera;

	// fRoll
    if (diks[DIK_LEFT]	&& 0x80)
		fRoll=-fTimedvarcamera;

    if (diks[DIK_RIGHT]	&& 0x80) 
		fRoll=+fTimedvarcamera;

	 // camera forward
    if (diks[DIK_HOME] &&0x80 )   // Key HOME     
	{
		vPos.x+=fspeed*vLook.x;
		vPos.y+=fspeed*vLook.y;
		vPos.z+=fspeed*vLook.z;
	}

	// camera back
    if (diks[DIK_END] &&0x80 )   // Key END     
	{
		vPos.x-=fspeed*vLook.x;
		vPos.y-=fspeed*vLook.y;
		vPos.z-=fspeed*vLook.z;
	}

	vLook = Normalize(vLook);
    vRight = CrossProduct( vUp, vLook); // Cross Produkt of the UP and LOOK Vector
	vRight = Normalize (vRight);
	vUp = CrossProduct (vLook, vRight); // Cross Produkt of the RIGHT and LOOK Vector
	vUp = Normalize(vUp);

    // Matrices for pitch, yaw and roll
	// This creates a rotation matrix around the viewers RIGHT vector. 
	D3DMATRIX matPitch, matYaw, matRoll;
    D3DUtil_SetRotationMatrix(matPitch, vRight, fPitch); 
	// Creates a rotation matrix around the viewers UP vector. 
    D3DUtil_SetRotationMatrix(matYaw, vUp, fYaw );
	// Creates a rotation matrix around the viewers LOOK vector. 
    D3DUtil_SetRotationMatrix(matRoll, vLook, fRoll);

	// now multiply these vectors with the matrices we've just created. 
	// First we rotate the LOOK & RIGHT Vectors about the UP Vector
    D3DMath_VectorMatrixMultiply(vLook , vLook, matYaw);
    D3DMath_VectorMatrixMultiply(vRight, vRight,matYaw);

	// And then we rotate the LOOK & UP Vectors about the RIGHT Vector
    D3DMath_VectorMatrixMultiply(vLook , vLook, matPitch);
    D3DMath_VectorMatrixMultiply(vUp, vUp, matPitch);

	// now rotate the RIGHT & UP Vectors about the LOOK Vector
    D3DMath_VectorMatrixMultiply(vRight, vRight, matRoll);
    D3DMath_VectorMatrixMultiply(vUp, vUp, matRoll);

	D3DMATRIX view=matWorld;
	D3DUtil_SetIdentityMatrix( view );// defined in d3dutil.h and d3dutil.cpp 
	view._11 = vRight.x; view._12 = vUp.x; view._13 = vLook.x;
	view._21 = vRight.y; view._22 = vUp.y; view._23 = vLook.y;
	view._31 = vRight.z; view._32 = vUp.z; view._33 = vLook.z;
	view._41 = - DotProduct( vPos, vRight ); // dot product defined in d3dtypes.h
	view._42 = - DotProduct( vPos, vUp );
	view._43 = - DotProduct( vPos, vLook );

	m_pd3dDevice->SetTransform(D3DTRANSFORMSTATE_VIEW,&view); 
  
    return S_OK;

}

//-----------------------------------------------------------------------------
// Name: DestroyInputDevice()
// Desc: Release the DirectInput device
//-----------------------------------------------------------------------------
VOID CMyD3DApplication::DestroyInputDevice()
{
	// keyboard
	if(g_Keyboard_pdidDevice2)
    {
        g_Keyboard_pdidDevice2->Unacquire();
        g_Keyboard_pdidDevice2->Release();
        g_Keyboard_pdidDevice2 = NULL;
    }
}

